@@ -24,9 +24,9 @@ prefix="/usr/pkg"
 # so that the unexpected situations with weird paths fall on the default
 # case of installed.
 script_path=$(cd "$(dirname "${0}")" && pwd)
-if test "${script_path}" = "/usr/pkgsrc/net/kea/work/kea-2.6.1/src/share/database/scripts/mysql"; then
+if test "${script_path}" = "/tmp/pkgbuild/net/kea/work.aarch64eb/kea-2.6.3/src/share/database/scripts/mysql"; then
     # shellcheck source=./src/bin/admin/admin-utils.sh.in
-    . "/usr/pkgsrc/net/kea/work/kea-2.6.1/src/bin/admin/admin-utils.sh"
+    . "/tmp/pkgbuild/net/kea/work.aarch64eb/kea-2.6.3/src/bin/admin/admin-utils.sh"
 else
     # shellcheck source=./src/bin/admin/admin-utils.sh.in
     . "${prefix}/share/kea/scripts/admin-utils.sh"
